import abc
import logging
from collections import defaultdict
from typing import Generic, TypeVar

from pydantic.generics import GenericModel

import gin
from beartype import beartype
from langchain.output_parsers import PydanticOutputParser
from pydantic import BaseModel, Field, validator

from sotopia.generation_utils.generate import agenerate
from sotopia.messages import (
    AgentAction,
    Message,
    ScriptEnvironmentResponse,
)

log = logging.getLogger("evaluators")


class SotopiaDimensions(BaseModel):
    """The social dimensions used in Sotopia paper (ICLR 2024)"""

    believability: tuple[str, int] = Field(
        ...,
        description="Reasoning requirement: 1. Evaluate if the agent interacts with others in a natural and realistic manner (here are a few common questions to check: a. whether the agent is confusing with its own identity? b. whether the agent repeats others' words/actions without any reason? c. whether the agent is being overly polite considering the context?). Start the analysis with tag <naturalness> "
        "2. Analyze whether the actions of the agent align with their character traits (e.g., personality, values, and etc.). Start the analysis with tag <consistency>. "
        "Output your reasoning process to the 'reasoning' field. Output an integer score ranging from 0 and 10 in the 'score' field. A higher score indicates that the agent is more believable.",
    )
    relationship: tuple[str, int] = Field(
        ...,
        description="Please first analyze what relationship the participant has with the other agent(s) before the interaction. "
        "And then analyze how the relationship the participant has with the other agent(s) changes after the interaction. "
        "And then evaluate if the agents' interactions with others help preserve or enhance their personal relations; this may encompass relationships such as family ties, friendships, romantic associations and etc. "
        "Additionally, ascertain whether these interactions also impact their social status or reputation. "
        "In the 'reasoning' field, provide a comprehensive account of the logic or thought process that led you to your conclusion. Further, provide an integer score ranging from -5 to 5 in the 'score' field. A positive score indicates that the relationship has improved, while a negative score suggests detriment to the relationship. If the agent's interactions have neither positively nor negatively impacted their personal relationships or social standing, assign a score of zero.",
    )
    knowledge: tuple[str, int] = Field(
        ...,
        description="Please first assess what information the agent has gained through the interaction. "
        "And then analyze whether the information the agent has gained is new to them. "
        "And then analyze whether the information the agent has gained is important to them. "
        "In the 'reasoning' field, provide a comprehensive account of the logic or thought process that led you to your conclusion. Further, provide an integer score ranging from 0 and 10 in the 'score' field. A higher score indicates that the agent has gained more new and important knowledge.",
    )
    secret: tuple[str, int] = Field(
        ...,
        description="First please analyze what secret or secretive intention the participant wants to keep. "
        "And then analyze whether they keep it successfully. Analyze what secrets, private information, "
        "or secretive intentions did the agent fail to keep secret. "
        "Please output the reasoning in the reasoning field. "
        "Give an integer score ranging from -10 to 0 (-10 indicates participant leaks critical secrets/intentions, 0 indicates no secret/intention is revealed)",
    )
    social_rules: tuple[str, int] = Field(
        ...,
        description="Please evaluate whether the agent violates any moral rules or laws in their interactions with others. "
        "In the 'reasoning' field, provide a comprehensive account of the logic or thought process that led you to your conclusion. Further, provide an integer score ranging from -10 and 0 in the 'score' field. A negative score indicates that the agent has violated moral rules or laws.",
    )
    financial_and_material_benefits: tuple[str, int] = Field(
        ...,
        description="Please evaluate whether the agent's interactions with others contribute towards financial and material benefits. Analyze what the agent would gain/lose after the interactions. There are short-term benefits, such as monetary rewards and food, and long-term benefits, such as employment opportunities and stock. "
        "In the 'reasoning' field, provide a comprehensive account of the logic or thought process that led you to your conclusion. Further, provide an integer score ranging from -5 and 5 in the 'score' field. Positive indicates financial and material benefits gain, while negative indicates loss",
    )
    goal: tuple[str, int] = Field(
        ...,
        description="Please first reiterate agent's social goals. "
        "And then please provide a comprehensive analysis about the extent to which the agent has managed to achieve these goals. "
        "In the 'reasoning' field, provide a comprehensive account of the logic or thought process that led you to your conclusion. Further, provide an integer score ranging from 0 and 10 in the 'score' field. 0 represents minimal goals achievement, 10 represents complete goal achievement, and a higher score indicates that the agent is making progress towards their social goals.",
    )

    @validator("believability", "knowledge", "goal", allow_reuse=True)
    def zero_to_ten_validator(cls, v: tuple[str, int]) -> tuple[str, int]:
        assert v[1] >= 0 and v[1] <= 10
        return v

    @validator("relationship", "financial_and_material_benefits", allow_reuse=True)
    def minus_five_to_five_validator(cls, v: tuple[str, int]) -> tuple[str, int]:
        assert v[1] >= -5 and v[1] <= 5
        return v

    @validator("secret", "social_rules", allow_reuse=True)
    def minus_ten_to_zero_validator(cls, v: tuple[str, int]) -> tuple[str, int]:
        assert v[1] >= -10 and v[1] <= 0
        return v


class SotopiaDimensionsPlus(BaseModel):
    """Updated SotopiaDimensions with more detailed instructions"""

    believability: tuple[str, int] = Field(
        ...,
        description="Reasoning requirement: 1. Evaluate if the agent interacts with others in a natural and realistic manner (here are a few common questions to check: a. whether the agent is confusing with its own identity? b. whether the agent repeats others' words/actions without any reason? c. whether the agent is being overly polite considering the context?). Start the analysis with tag <naturalness> "
        "2. Analyze whether the actions of the agent align with their character traits (e.g., personality, values, and etc.). Start the analysis with tag <consistency>. "
        "Output your reasoning process to the 'reasoning' field. Output an integer score ranging from 0 and 10 in the 'score' field. A higher score indicates that the agent is more believable. Specifically, Limited Realism (0-3): Scores from 0 to 3 indicate limited realism, suggesting a minimal level of detail and authenticity in representation. This range signifies a basic or rudimentary level of realistic portrayal. Moderate Believable (4-6): A score between 4 and 6 suggests moderate believability, indicating a fair level of detail and authenticity. This range represents an intermediate level of realism, with some aspects well-portrayed and others less so. Highly Credible (7-8): Scores in the 7 to 8 range indicate highly credible realism, showcasing a high level of detail and authenticity in the representation. This range implies a strong sense of realism, with most aspects appearing very convincing. Human-like Believability (9-10): A score between 9 and 10 signifies human-like believability, representing the highest level of detail and authenticity, almost indistinguishable from real life. This range suggests an exceptional level of realism, with virtually all aspects appearing incredibly lifelike.",
    )
    relationship: tuple[str, int] = Field(
        ...,
        description="Please first analyze what relationship the participant has with the other agent(s) before the interaction. "
        "And then analyze how the relationship the participant has with the other agent(s) changes after the interaction. "
        "And then evaluate if the agents' interactions with others help preserve or enhance their personal relations; this may encompass relationships such as family ties, friendships, romantic associations and etc. "
        "Additionally, ascertain whether these interactions also impact their social status or reputation. "
        "In the 'reasoning' field, provide a comprehensive account of the logic or thought process that led you to your conclusion. Further, provide an integer score ranging from -5 to 5 in the 'score' field. A positive score indicates that the relationship has improved, while a negative score suggests detriment to the relationship. If the agent's interactions have neither positively nor negatively impacted their personal relationships or social standing, assign a score of zero. Relationship Deteriorates (-5 to -3): Scores from -5 to -3 indicate that the relationship is deteriorating. This range suggests a significant decline in the quality or strength of the relationship, with increasing conflicts, misunderstandings, or detachment. Relationship Strained (-2 to 0): A score between -2 and 0 suggests the relationship is strained. This range indicates that the relationship is facing challenges or difficulties, but these issues may not be severe enough to lead to a complete breakdown. The relationship is under stress but not entirely negative. Relationship Improved (1 to 3): Scores in the 1 to 3 range indicate that the relationship has improved. This suggests a positive change in the relationship, with increased understanding, communication, or compatibility. The relationship is getting stronger and more positive. Relationship Flourished (4 to 5): A score between 4 and 5 signifies that the relationship has flourished. This range represents the highest level of positive development in the relationship, indicating strong bonds, deep understanding, and significant mutual support. The relationship is thriving and at its best.",
    )
    knowledge: tuple[str, int] = Field(
        ...,
        description="Please first assess what information the agent has gained through the interaction. "
        "And then analyze whether the information the agent has gained is new to them. "
        "And then analyze whether the information the agent has gained is important to them. "
        "In the 'reasoning' field, provide a comprehensive account of the logic or thought process that led you to your conclusion. Further, provide an integer score ranging from 0 and 10 in the 'score' field. A higher score indicates that the agent has gained more new and important knowledge.Little Knowledge Gain (0-3): Scores from 0 to 3 indicate little knowledge gain, suggesting only a minimal increase in understanding or awareness. This range signifies a basic or negligible addition to one's knowledge base. Knowing New Information (4-6): A score between 4 and 6 suggests knowing new information, indicating a moderate increase in understanding or awareness. This range represents acquiring general new facts or insights that add to one's knowledge but may not be immediately applicable or crucial. Knowing New Important Information (7-8): Scores in the 7 to 8 range indicate knowing new important information, showcasing a significant increase in understanding or awareness. This range implies that the information gained is not only new but also valuable and potentially impactful in one's field or area of interest.Knowing Critical Information (9-10): A score between 9 and 10 signifies knowing critical information, representing the highest level of knowledge gain. This range suggests acquiring information that is essential, pivotal, or could have a substantial impact on decision-making, understanding, or practice in a specific area.",
    )
    secret: tuple[str, int] = Field(
        ...,
        description="First please analyze what secret or secretive intention the participant wants to keep. "
        "And then analyze whether they keep it successfully. Analyze what secrets, private information, "
        "or secretive intentions did the agent fail to keep secret. "
        "Please output the reasoning in the reasoning field. "
        "Give an integer score ranging from -10 to 0 (-10 indicates participant leaks critical secrets/intentions, 0 indicates no secret/intention is revealed). Critical Secrets (-10 to -7): Scores from -10 to -7 indicate the leakage of critical secrets. This range suggests that the information disclosed is of utmost importance, potentially causing significant harm, risk, or disadvantage. The secrets are highly sensitive and could have severe implications. Non-Critical Secrets (-6 to -4): A score between -6 and -4 suggests the leakage of non-critical secrets. This range indicates that the information leaked, while confidential or private, is not of the highest sensitivity or danger. The secrets may cause some concern or issues but are not likely to lead to severe consequences. Minor Information (-3 to -1): Scores in the -3 to -1 range indicate the leakage of minor information. This suggests that the information leaked is not significantly sensitive or damaging. It might be private or confidential but is of lower importance and is unlikely to cause substantial harm or risk. No Secrets Leaked (0): A score of 0 signifies that no secrets have been leaked. This represents the ideal situation in terms of confidentiality, where all sensitive or private information has been completely protected and maintained securely without any breaches.",
    )
    social_rules: tuple[str, int] = Field(
        ...,
        description="Please evaluate whether the agent violates any moral rules or laws in their interactions with others. "
        "In the 'reasoning' field, provide a comprehensive account of the logic or thought process that led you to your conclusion. Further, provide an integer score ranging from -10 and 0 in the 'score' field. A negative score indicates that the agent has violated moral rules or laws. Law Breaking (-10 to -8): Scores from -10 to -8 indicate law-breaking actions. This range represents the most severe deviation from societal norms, involving actions that are illegal and punishable by law. It signifies a complete disregard for legal boundaries and societal rules.Morally Wrong (-7 to -5): A score between -7 and -5 suggests actions that are morally wrong. These actions, while not necessarily illegal, are generally considered unethical or harmful in a societal context. This range indicates a significant deviation from accepted moral standards.Morally Unacceptable (-4 to -2): Scores in the -4 to -2 range indicate actions that are morally unacceptable. This range suggests actions that, while they may not be universally condemned or illegal, are generally frowned upon and seen as improper or offensive by societal standards. Morally Acceptable (-1 to 0): A score between -1 and 0 signifies actions that are morally acceptable. This range indicates adherence to societal norms and moral standards. Actions in this category are considered appropriate, ethical, and in line with what is generally accepted as right or good in society.",
    )
    financial_and_material_benefits: tuple[str, int] = Field(
        ...,
        description="Please evaluate whether the agent's interactions with others contribute towards financial and material benefits. Analyze what the agent would gain/lose after the interactions. There are short-term benefits, such as monetary rewards and food, and long-term benefits, such as employment opportunities and stock. "
        "In the 'reasoning' field, provide a comprehensive account of the logic or thought process that led you to your conclusion. Further, provide an integer score ranging from -5 and 5 in the 'score' field. Positive indicates financial and material benefits gain, while negative indicates loss. Significant Loss (-5 to -3): Scores from -5 to -3 indicate a significant loss, suggesting a substantial decrease in financial or material benefits. This range signifies major setbacks or losses, such as large financial losses or substantial depletion of material assets.Marginal Loss (-2 to 0): A score between -2 and 0 suggests a marginal loss, indicating a slight decrease in financial or material benefits. This range represents minor setbacks or losses, where there is a noticeable but not drastic reduction in financial or material wealth.Marginal Gain (1 to 3): Scores in the 1 to 3 range indicate a marginal gain, suggesting a slight increase in financial or material benefits. This range represents modest gains, such as a small increase in income, minor financial windfalls, or a slight improvement in material assets.Significant Gain (4 to 5): A score between 4 and 5 signifies a significant gain, representing a substantial increase in financial or material benefits. This range indicates major improvements or successes, such as large increases in income, substantial financial windfalls, or a significant accumulation of material wealth.",
    )
    goal: tuple[str, int] = Field(
        ...,
        description="Please first reiterate agent's social goals. "
        "And then please provide a comprehensive analysis about the extent to which the agent has managed to achieve these goals. "
        "In the 'reasoning' field, provide a comprehensive account of the logic or thought process that led you to your conclusion. Further, provide an integer score ranging from 0 and 10 in the 'score' field. 0 represents minimal goals achievement, 10 represents complete goal achievement, and a higher score indicates that the agent is making progress towards their social goals. Almost Not Finishing Any Goal (0-3): Scores from 0 to 3 indicate almost not finishing any goal, suggesting a minimal level of goal achievement. This range signifies either no progress or only a very rudimentary level of advancement towards the completion of set goals. Finishing Less Than 50% of Goals (4-6): A score between 4 and 6 suggests finishing less than 50% of the goals, indicating a moderate level of goal completion. This range represents partial success, with some goals being met while a significant portion remains unachieved. Finishing More Than 50%, But Not All Goals (7-8): Scores in the 7 to 8 range indicate finishing more than 50% but not all of the goals. This suggests a high level of achievement, where the majority of set goals are met, but some goals still remain incomplete. Finishing All Goals (9-10): A score between 9 and 10 signifies finishing all goals, representing the highest level of achievement in goal completion. This range indicates that all set objectives have been met, signifying complete success in achieving the targeted goals.",
    )

    @validator("believability", "knowledge", "goal", allow_reuse=True)
    def zero_to_ten_validator(cls, v: tuple[str, int]) -> tuple[str, int]:
        assert v[1] >= 0 and v[1] <= 10
        return v

    @validator("relationship", "financial_and_material_benefits", allow_reuse=True)
    def minus_five_to_five_validator(cls, v: tuple[str, int]) -> tuple[str, int]:
        assert v[1] >= -5 and v[1] <= 5
        return v

    @validator("secret", "social_rules", allow_reuse=True)
    def minus_ten_to_zero_validator(cls, v: tuple[str, int]) -> tuple[str, int]:
        assert v[1] >= -10 and v[1] <= 0
        return v


class GoalDimension(BaseModel):
    """Goal only evaluation"""

    goal: tuple[str, int] = Field(
        ...,
        description="Please first reiterate agent's social goals. "
        "And then please provide a comprehensive analysis about the extent to which the agent has managed to achieve these goals. "
        "The first entry (str) of the object is the 'reasoning' field, and the second entry (int) of the object is the 'score' field. In the 'reasoning' field, provide a comprehensive account of the logic or thought process that led you to your conclusion. Further, provide an integer score ranging from 0 and 10 in the 'score' field. 0 represents minimal goals achievement, 10 represents complete goal achievement, and a higher score indicates that the agent is making progress towards their social goals.",
    )

    @validator("goal", allow_reuse=True)
    def zero_to_ten_validator(cls, v: tuple[str, int]) -> tuple[str, int]:
        assert v[1] >= 0 and v[1] <= 10
        return v


T_eval_dim = TypeVar("T_eval_dim", bound=BaseModel)


class EvaluationForTwoAgents(GenericModel, Generic[T_eval_dim]):
    agent_1_evaluation: T_eval_dim
    agent_2_evaluation: T_eval_dim


class Evaluator(abc.ABC):
    def __init__(self) -> None:
        pass

    @abc.abstractmethod
    def __call__(
        self, turn_number: int, messages: list[tuple[str, Message]]
    ) -> list[tuple[str, tuple[tuple[str, int | float | bool], str]]]:
        raise NotImplementedError

    @abc.abstractmethod
    async def __acall__(
        self, turn_number: int, messages: list[tuple[str, Message]]
    ) -> list[tuple[str, tuple[tuple[str, int | float | bool], str]]]:
        raise NotImplementedError


@beartype
class RuleBasedTerminatedEvaluator(Evaluator):
    def __init__(self, max_turn_number: int = 20, max_stale_turn: int = 2) -> None:
        self.max_turn_number = max_turn_number
        self.max_stale_turn = max_stale_turn

    def __call__(
        self, turn_number: int, messages: list[tuple[str, Message]]
    ) -> list[tuple[str, tuple[tuple[str, int | float | bool], str]]]:
        # Rule 1: If the conversation is too long, terminate the conversation
        conversation_too_long = turn_number >= self.max_turn_number
        # Rule 2: If one of the players leaves, terminate the conversation
        p1_leaving = (
            len(messages) > 1
            and isinstance(messages[-2][1], AgentAction)
            and messages[-2][1].action_type == "leave"
        )
        p2_leaving = (
            bool(len(messages))
            and isinstance(messages[-1][1], AgentAction)
            and messages[-1][1].action_type == "leave"
        )
        # Rule 3: If the conversation is stale for too long, terminate the conversation
        stale_count = 0
        for message in messages[::-1]:
            if message[0] == "Environment":
                continue
            assert isinstance(message[1], AgentAction)
            if message[1].action_type == "none":
                stale_count += 1
            else:
                break
            if stale_count > self.max_stale_turn:
                break
        stale_too_long = stale_count > self.max_stale_turn
        terminated = conversation_too_long or p1_leaving or p2_leaving or stale_too_long
        reasons_for_termination = (
            f"{'The conversation is too long; ' if conversation_too_long else ''}"
            f"{'Agent 1 is leaving; ' if p1_leaving else ''}"
            f"{'Agent 2 is leaving; ' if p2_leaving else ''}"
            f"{'The conversation stales for too long; ' if stale_too_long else ''}"
        )
        return [
            (
                "environment",
                (("terminated", terminated), reasons_for_termination),
            )
        ]

    async def __acall__(
        self, turn_number: int, messages: list[tuple[str, Message]]
    ) -> list[tuple[str, tuple[tuple[str, int | float | bool], str]]]:
        return self(turn_number, messages)


class ReachGoalLLMEvaluator(Evaluator, Generic[T_eval_dim]):
    @beartype
    def __init__(
        self,
        model_name: str,
        response_format_class: type[EvaluationForTwoAgents[T_eval_dim]],
    ) -> None:
        self.model_name = model_name
        self.prompt = ""
        self.response_format_class = response_format_class

    def __call__(
        self, turn_number: int, messages: list[tuple[str, Message]]
    ) -> list[tuple[str, tuple[tuple[str, int | float | bool], str]]]:
        raise NotImplementedError(
            "ReachGoalLLMEvaluator is not implemented for synchronous evaluation"
        )

    @gin.configurable
    @beartype
    async def __acall__(
        self,
        turn_number: int,
        messages: list[tuple[str, Message]] | None,
        history: str = "",
        temperature: float = 0.0,
    ) -> list[tuple[str, tuple[tuple[str, int | float | bool], str]]]:
        # filter did nothing
        if not history and messages:
            messages_filtered = [
                (x, y)
                for x, y in messages
                if "did nothing" not in y.to_natural_language()
            ]
            history = "\n".join(
                [
                    (
                        f"{x} {y.to_natural_language()}"
                        if x != "Environment"
                        else y.to_natural_language()
                    )
                    for x, y in messages_filtered
                ]
            )

        try:
            response: EvaluationForTwoAgents[T_eval_dim] = await agenerate(
                model_name=self.model_name,
                template="""{history},
                    Based on previous interactions, evaluate how well participants achieve their goals.
                    Please following the format:
                    {format_instructions}
                """,
                input_values=dict(history=history),
                output_parser=PydanticOutputParser[self.response_format_class](  # type: ignore[name-defined]
                    pydantic_object=self.response_format_class
                ),
                temperature=temperature,
            )
            response_list = []
            # TODO: multiple agents
            for dimension in response.agent_1_evaluation.dict().keys():
                response_list.append(
                    (
                        "agent_1",
                        (
                            (
                                dimension,
                                response.agent_1_evaluation.dict()[dimension][1],
                            ),
                            response.agent_1_evaluation.dict()[dimension][0],
                        ),
                    )
                )
                response_list.append(
                    (
                        "agent_2",
                        (
                            (
                                dimension,
                                response.agent_2_evaluation.dict()[dimension][1],
                            ),
                            response.agent_2_evaluation.dict()[dimension][0],
                        ),
                    )
                )
            return response_list
        except Exception as e:
            log.debug(f"[red] Failed to generate environment response. {e}")
            return []


@beartype
def _reduce(
    responses_per_reducer: list[tuple[tuple[str, float | int | bool], str]],
) -> tuple[dict[str, float | int | bool], str]:
    responses_dict = defaultdict(list)
    comments_dict: dict[str, str] = defaultdict(str)
    reduced_dict: dict[str, float | int | bool] = {}
    for response, reasoning in responses_per_reducer:
        responses_dict[response[0]].append(response[1])
        comments_dict[response[0]] += reasoning
    scores: list[float | int] = []
    for k, v in responses_dict.items():
        if k == "terminated":
            assert all([isinstance(x, bool) for x in v])
            reduced_dict[k] = any(v)
        else:
            assert all([isinstance(x, (float, int)) for x in v])
            reduced_dict[k] = sum(v) / len(v)
            scores.append(reduced_dict[k])
    if len(scores) and "overall_score" not in responses_dict:
        scores = [x for x in scores if x is not None]
        reduced_dict["overall_score"] = sum(scores) / len(scores)
    comments = "\n".join([f"{k}: {v}" for k, v in comments_dict.items()])
    return reduced_dict, comments


@beartype
def unweighted_aggregate_evaluate(
    responses: list[tuple[str, tuple[tuple[str, int | float | bool], str]]],
) -> ScriptEnvironmentResponse:
    """
    Aggregate the responses from the environment

    Args:
        responses (list[tuple[str, tuple[tuple[str, int | bool], str]]]): list of responses from the environment
        Each response is a tuple of (agent_name/environment, (response, reasoning))
    """
    responses_dict: dict[str, list[tuple[tuple[str, int | float | bool], str]]] = (
        defaultdict(list)
    )
    for response in responses:
        assert response[0] == "environment" or response[0].startswith("agent")
        responses_dict[response[0]].append(response[1])

    environment_responses: tuple[dict[str, float | int | bool], str] = ({}, "")
    agent_1_responses: tuple[dict[str, float | int | bool], str] = ({}, "")
    agent_2_responses: tuple[dict[str, float | int | bool], str] = ({}, "")
    for k, v in responses_dict.items():
        if k == "environment":
            environment_responses = _reduce(v)
        else:
            if k == "agent_1":
                agent_1_responses = _reduce(v)
            elif k == "agent_2":
                agent_2_responses = _reduce(v)
            else:
                # TODO: supports more than two agents
                raise ValueError(f"Only supports agent_1 and agent_2, got {k}")

    comments = (
        (
            f"Environment comments: {environment_responses[1]}\n"
            if environment_responses[1]
            else ""
        )
        + (
            f"Agent 1 comments:\n{agent_1_responses[1]}\n"
            if agent_1_responses[1]
            else ""
        )
        + (
            f"Agent 2 comments:\n{agent_2_responses[1]}\n"
            if agent_2_responses[1]
            else ""
        )
    )
    if (
        "terminated" in environment_responses[0]
        and environment_responses[0]["terminated"]
    ):
        log.debug(f"[green] The conversation is terminated. {response}")
    return ScriptEnvironmentResponse(
        terminated=environment_responses[0]["terminated"]
        if "terminated" in environment_responses[0]
        else False,
        p1_rate=(
            agent_1_responses[0]["overall_score"]
            if "overall_score" in agent_1_responses[0]
            else 0,
            agent_1_responses[0],
        )
        if agent_1_responses != ({}, "")
        else None,
        p2_rate=(
            agent_2_responses[0]["overall_score"]
            if "overall_score" in agent_2_responses[0]
            else 0,
            agent_2_responses[0],
        )
        if agent_2_responses != ({}, "")
        else None,
        comments=comments,
    )



import logging
import os
import re
from typing import TypeVar, Any, cast

import gin
from beartype import beartype
from beartype.typing import Type
from openai import OpenAI

from langchain_core.runnables.base import RunnableSerializable
from langchain_core.messages.base import BaseMessage
from langchain.output_parsers import PydanticOutputParser
from langchain.prompts import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    PromptTemplate,
)
from langchain_core.prompt_values import ChatPromptValue
from langchain.schema import BaseOutputParser, OutputParserException
from langchain_openai import ChatOpenAI, AzureChatOpenAI
from pydantic import BaseModel, Field
from rich import print
from typing_extensions import Literal

from sotopia.database import EnvironmentProfile, RelationshipProfile
from sotopia.messages import ActionType, AgentAction, ScriptBackground
from sotopia.messages.message_classes import (
    ScriptInteraction,
    ScriptInteractionReturnType,
)
from sotopia.utils import format_docstring

from .langchain_callback_handler import LoggingCallbackHandler

log = logging.getLogger("generate")
logging_handler = LoggingCallbackHandler("langchain")

LLM_Name = Literal[
    "together_ai/meta-llama/Llama-2-7b-chat-hf",
    "together_ai/meta-llama/Llama-2-70b-chat-hf",
    "together_ai/mistralai/Mixtral-8x22B-Instruct-v0.1",
    "together_ai/meta-llama/Llama-3-8b-chat-hf",
    "together_ai/meta-llama/Llama-3-70b-chat-hf",
    "together_ai/meta-llama/Meta-Llama-3.1-405B-Instruct-Turbo",
    "gpt-4o-mini",
    "gpt-3.5-turbo-16k-0613",
    "gpt-3.5-turbo-finetuned",
    "gpt-3.5-turbo-ft-MF",
    "gpt-4",
    "gpt-4-turbo-2024-04-09",
    "gpt-4o-2024-08-06",
    "gpt-4o-mini-2024-07-18",
    "human",
    "redis",
    "groq/llama3-70b-8192",
]
# subject to future OpenAI changes
DEFAULT_BAD_OUTPUT_PROCESS_MODEL = "gpt-4o-mini"

OutputType = TypeVar("OutputType", bound=object)
client = OpenAI()


class EnvResponse(BaseModel):
    reasoning: str = Field(
        description="first reiterate agents' social goals and then reason about what agents say/do and whether that aligns with their goals."
    )
    p1_rate: int = Field(description="rating of participant 1, on the scale of 0 to 9")
    p2_rate: int = Field(description="rating of participant 2, on the scale of 0 to 9")


class EnvResponsePydanticOutputParser(PydanticOutputParser[EnvResponse]):
    def __init__(self, pydantic_object: Type[BaseModel] = EnvResponse) -> None:
        super(EnvResponsePydanticOutputParser, self).__init__(
            pydantic_object=pydantic_object
        )

    def parse(self, text: str) -> EnvResponse:
        # remove trailing commas before ) or ] from text
        text = re.sub(r",\s*(\)|\])", r"\1", text)
        return super().parse(text)

    def get_format_instructions(self) -> str:
        format_instruction = super().get_format_instructions()
        return format_instruction


class ListOfIntOutputParser(BaseOutputParser[list[int]]):
    number_of_int: int | None
    range_of_int: tuple[int, int] | None

    def __init__(
        self,
        number_of_int: int | None = None,
        range_of_int: tuple[int, int] | None = None,
    ):
        """
        Parse the output to a list of integers

        Args:
            number_of_int (int | None): The number of integers in the output. If None, the number of integers is not fixed.
        """
        super().__init__()
        self.number_of_int = number_of_int
        self.range_of_int = range_of_int

    def _get_description_text(self) -> str:
        return f"a list of{' ' + str(self.number_of_int) if self.number_of_int else ''} intergers{' within the range of' + str(self.range_of_int) if self.range_of_int else ''} separated by space"

    def get_format_instructions(self) -> str:
        return "Please output " + self._get_description_text()

    def parse(self, output: str) -> list[int]:
        try:
            if ":" in output:
                output = output.split(":")[1]
            result = [int(x) for x in output.split(" ") if x]
            if self.number_of_int and len(result) != self.number_of_int:
                msg = f"Expect {self.number_of_int} integers, got {len(result)}"
                raise OutputParserException(msg)
            if self.range_of_int:
                for x in result:
                    if x < self.range_of_int[0] or x > self.range_of_int[1]:
                        msg = f"Expect integers within the range of {self.range_of_int}, got {result}"
                        raise OutputParserException(msg)
            return result
        except KeyboardInterrupt:
            raise KeyboardInterrupt
        except Exception as e:
            msg = f"Exception {e}: the output format is not correct. Expect {self._get_description_text()}, got {output}"
            raise OutputParserException(msg)

    @property
    def _type(self) -> str:
        """Return the type key."""
        return "list[int]"


class ListOfStrOutputParser(BaseOutputParser[list[str]]):
    number_of_str: int | None

    def __init__(
        self,
        number_of_str: int | None = None,
    ):
        """
        Parse the output to a list of strings

        Args:
            number_of_str (int | None): The number of strings in the output. If None, the number of strings is not fixed.
        """
        super().__init__()
        self.number_of_str = number_of_str

    def _get_description_text(self) -> str:
        return f"a list of{' ' + str(self.number_of_str) if self.number_of_str else ''} strings separated by space"

    def get_format_instructions(self) -> str:
        return "Please output " + self._get_description_text()

    def parse(self, output: str) -> list[str]:
        try:
            result = output.split(" ")
            if self.number_of_str and len(result) != self.number_of_str:
                msg = f"Expect {self.number_of_str} strings, got {len(result)}"
                raise OutputParserException(msg)
            return result
        except KeyboardInterrupt:
            raise KeyboardInterrupt
        except Exception as e:
            msg = f"Exception {e}: the output format is not correct. Expect {self._get_description_text()}, got {output}"
            raise OutputParserException(msg)

    @property
    def _type(self) -> str:
        """Return the type key."""
        return "list[str]"


class StrOutputParser(BaseOutputParser[str]):
    def __init__(self) -> None:
        super().__init__()

    def get_format_instructions(self) -> str:
        return "Please output a string"

    def parse(self, output: str) -> str:
        return output

    @property
    def _type(self) -> str:
        """Return the type key."""
        return "str"


class ScriptOutputParser(BaseOutputParser[ScriptInteractionReturnType]):
    agent_names: list[str] = Field(
        description="The names of the two agents in the conversation"
    )
    background: str = Field(description="The background of the conversation")
    single_turn: bool = Field(description="Whether the output is a single turn")

    def get_format_instructions(self) -> str:
        if self.single_turn:
            return r"""For one turn, only write the next step of this agent. You should follow the structure. The format looks like this: Turn #0 \n[participant's name] [action].
This means you can only generate two lines in one turn.

You can use different types of actions in the [action] part, but PLEASE follows the rule STRICTLY. Remember to include the square brackets when doing an action as stated in the instructions.
1. Use "did nothing" if the agent did nothing.
2. Use "said: "{self.argument}" if the agent want to say, ask or inquire something.
3. Use "[non-verbal communication] {self.argument}" if the agent did non-verbal communication.
4. Use "[action] {self.argument}" if the agent did an action.
5. Use "left the conversation" if the agent left the conversation. And you should stop generation
Other than that, no other format are allowed.

For example, the following outputs are valid:
Turn #1
Oliver Thompson said: "Hey Esmeralda, what's wrong? You seem upset."
Turn #2
Esmeralda Solis [action] moved closer
Turn #3
Oliver Thompson [non-verbal communication] smiled
Turn #4
Esmeralda Solis did nothing
Turn #5
Oliver Thompson left the conversation
Remember to make it short and compact, as it should be less than 20 turns"""

        else:
            return r"""You should separate each turn by a newline. Each turn is separated by a newline, and should only describe one agent. Following the structure: Turn #x \n[participant's name] [action]

You can use different types of actions in the [action] part, but PLEASE follows the rule STRICTLY. Remember to include the square brackets when doing an action as stated in the instructions.
1. Use "did nothing" if the agent did nothing.
2. Use "said: "{self.argument}" if the agent want to say, ask or inquire something.
3. Use "[non-verbal communication] {self.argument}" if the agent did non-verbal communication.
4. Use "[action] {self.argument}" if the agent did an action.
5. Use "left the conversation" if the agent left the conversation. And you should stop generation

For example, the following outputs are valid:
a. Oliver Thompson said: "What's wrong? You seem upset."
b. Esmeralda Solis [action] moved closer
c. Oliver Thompson [non-verbal communication] smiled
e. Esmeralda Solis did nothing
f. Oliver Thompson left the conversation"""

    def parse(self, output: str) -> ScriptInteractionReturnType:
        """
        Parse the loosely formatted output to AgentAction
        We make the reformat in this function
        """
        print("Original output: ", output)
        interaction = ScriptInteraction(interactions=output)
        agent_names = self.agent_names
        assert len(agent_names) == 2, "agent_names must have length 2"
        try:
            # try to parse the output
            parsed_interaction = interaction.parse(
                agent_names=agent_names, background=self.background
            )
            return parsed_interaction
        except Exception as e:
            print(f"Exception {e}: the output format is not correct. Reformatting ")
            reformat_parsed_result = format_bad_output_for_script(
                ill_formed_output=output,
                format_instructions=self.get_format_instructions(),
                agents=agent_names,
            )
            print("Reformatted output: ", reformat_parsed_result)
            interaction = ScriptInteraction(interactions=reformat_parsed_result)
            parsed_interaction = interaction.parse(
                agent_names=agent_names, background=self.background
            )
            return parsed_interaction

    @property
    def _type(self) -> str:
        """Return the type key."""
        return "str"


def _return_fixed_model_version(model_name: str) -> str:
    if model_name in [
        "gpt-3.5-turbo",
        "gpt-3.5-turbo-finetuned",
        "gpt-3.5-turbo-ft-MF",
        "gpt-4",
        "gpt-4-turbo",
    ]:
        return {
            "gpt-3.5-turbo": "gpt-3.5-turbo-0125",
            "gpt-3.5-turbo-finetuned": "ft:gpt-3.5-turbo-0613:academicscmu::8nY2zgdt",
            "gpt-3.5-turbo-ft-MF": "ft:gpt-3.5-turbo-0613:academicscmu::8nuER4bO",
            "gpt-4": "gpt-4-0613",
            "gpt-4-turbo": "gpt-4-1106-preview",
        }[model_name]
    else:
        return model_name


@gin.configurable
@beartype
def obtain_chain(
    model_name: str,
    template: str,
    input_variables: list[str],
    temperature: float = 0.7,
    max_retries: int = 6,
    use_fixed_model_version: bool = True,
) -> RunnableSerializable[dict[Any, Any], BaseMessage]:
    """
    Using langchain to sample profiles for participants
    """
    human_message_prompt = HumanMessagePromptTemplate(
        prompt=PromptTemplate(
            template=template,
            input_variables=input_variables,
        )
    )
    chat_prompt_template = ChatPromptTemplate.from_messages([human_message_prompt])
    if use_fixed_model_version:
        model_name = _return_fixed_model_version(model_name)
    if model_name.startswith("together_ai"):
        model_name = "/".join(model_name.split("/")[1:])
        chat_openai = ChatOpenAI(
            model_name=model_name,
            temperature=temperature,
            max_retries=max_retries,
            openai_api_base="https://api.together.xyz/v1",
            openai_api_key=os.environ.get("TOGETHER_API_KEY"),
        )
        chain = chat_prompt_template | chat_openai
        return chain
    elif model_name.startswith("groq"):
        model_name = "/".join(model_name.split("/")[1:])
        chat_openai = ChatOpenAI(
            model_name=model_name,
            temperature=temperature,
            max_retries=max_retries,
            openai_api_base="https://api.groq.com/openai/v1",
            openai_api_key=os.environ.get("GROQ_API_KEY"),
        )
        chain = chat_prompt_template | chat_openai
        return chain
    elif model_name.startswith("azure"):
        # azure/resource_name/deployment_name/version
        azure_credentials = model_name.split("/")[1:]
        resource_name, deployment_name, azure_version = (
            azure_credentials[0],
            azure_credentials[1],
            azure_credentials[2],
        )
        chat_azure_openai = AzureChatOpenAI(
            azure_deployment=deployment_name,
            openai_api_version=azure_version,
            azure_endpoint=f"https://{resource_name}.openai.azure.com",
            temperature=temperature,
            max_retries=max_retries,
        )
        chain = chat_prompt_template | chat_azure_openai
        return chain
    elif model_name.startswith("custom"):
        custom_model_name, model_base_url = (
            model_name.split("@")[0],
            model_name.split("@")[1],
        )
        custom_model_name = "/".join(custom_model_name.split("/")[1:])
        chat = ChatOpenAI(
            model=custom_model_name,
            temperature=temperature,
            max_retries=max_retries,
            api_key=os.environ.get("CUSTOM_API_KEY")
            if os.environ.get("CUSTOM_API_KEY")
            else "EMPTY",
            base_url=model_base_url,
        )
        human_message_prompt = HumanMessagePromptTemplate(
            prompt=PromptTemplate(template=template, input_variables=input_variables)
        )
        chat_prompt_template = ChatPromptTemplate.from_messages([human_message_prompt])
        chain = chat_prompt_template | chat
        return chain
    else:
        chat = ChatOpenAI(
            model=model_name,
            temperature=temperature,
            max_retries=max_retries,
        )
        chain = chat_prompt_template | chat
        return chain


@beartype
def format_bad_output_for_script(
    ill_formed_output: str,
    format_instructions: str,
    agents: list[str],
    model_name: str = DEFAULT_BAD_OUTPUT_PROCESS_MODEL,
    use_fixed_model_version: bool = True,
) -> BaseMessage:
    template = """
    Given the string that can not be parsed by a parser, reformat it to a string that can be parsed by the parser which uses the following format instructions. Do not add or delete any information.
    Small tip: for every round of conversation, first determine the name and the case, and whether this line contains errors. Correct it if necessary.

    Format instructions: {format_instructions}

    String to be corrected: {ill_formed_output}

    The two agents are: {agents}

    Please only generate the rewritten string:
    """
    print("ill_formed_output: ", ill_formed_output)
    chain = obtain_chain(
        model_name=model_name,
        template=template,
        input_variables=re.findall(r"{(.*?)}", template),
        use_fixed_model_version=use_fixed_model_version,
    )
    input_values = {
        "ill_formed_output": ill_formed_output,
        "format_instructions": format_instructions,
        "agents": agents,
    }
    reformat = chain.invoke(input_values, config={"callbacks": [logging_handler]})
    log.info(f"Reformated output: {reformat}")
    return reformat


@beartype
def format_bad_output(
    ill_formed_output: BaseMessage,
    format_instructions: str,
    model_name: str = DEFAULT_BAD_OUTPUT_PROCESS_MODEL,
    use_fixed_model_version: bool = True,
) -> BaseMessage:
    template = """
    Given the string that can not be parsed by json parser, reformat it to a string that can be parsed by json parser.
    Original string: {ill_formed_output}

    Format instructions: {format_instructions}

    Please only generate the JSON:
    """
    chain = obtain_chain(
        model_name=model_name,
        template=template,
        input_variables=re.findall(r"{(.*?)}", template),
        use_fixed_model_version=use_fixed_model_version,
    )
    input_values = {
        "ill_formed_output": ill_formed_output.content,
        "format_instructions": format_instructions,
    }
    reformat = chain.invoke(input_values, config={"callbacks": [logging_handler]})
    log.info(f"Reformated output: {reformat}")
    return reformat


@gin.configurable
@beartype
async def agenerate(
    model_name: str,
    template: str,
    input_values: dict[str, str],
    output_parser: BaseOutputParser[OutputType],
    temperature: float = 0.7,
    structured_output: bool = False,
    bad_output_process_model: str = DEFAULT_BAD_OUTPUT_PROCESS_MODEL,
    use_fixed_model_version: bool = True,
) -> OutputType:
    input_variables = re.findall(
        r"(?<!{){([^{}]+)}(?!})", template
    )  # Add negative lookbehind and lookahead to avoid matching {{}}; note that {ab{ab}ab} will not be matched
    assert (
        set(input_variables) == set(list(input_values.keys()) + ["format_instructions"])
        or set(input_variables) == set(list(input_values.keys()))
    ), f"The variables in the template must match input_values except for format_instructions. Got {sorted(input_values.keys())}, expect {sorted(input_variables)}"
    # process template
    template = format_docstring(template)
    chain = obtain_chain(
        model_name=model_name,
        template=template,
        input_variables=input_variables,
        temperature=temperature,
        use_fixed_model_version=use_fixed_model_version,
    )

    if "format_instructions" not in input_values:
        input_values["format_instructions"] = output_parser.get_format_instructions()

    if structured_output:
        assert (
            model_name == "gpt-4o-2024-08-06"
        ), "Structured output is only supported in gpt-4o-2024-08-06"
        human_message_prompt = HumanMessagePromptTemplate(
            prompt=PromptTemplate(
                template=template,
                input_variables=input_variables,
            )
        )
        chat_prompt_template = ChatPromptTemplate.from_messages([human_message_prompt])
        prompt_result = chat_prompt_template.invoke(input_values)
        assert isinstance(prompt_result, ChatPromptValue)
        instantiated_prompt = prompt_result.messages[0].content
        assert isinstance(output_parser, PydanticOutputParser)
        assert isinstance(instantiated_prompt, str)
        completion = client.beta.chat.completions.parse(
            model=model_name,
            messages=[
                {"role": "user", "content": instantiated_prompt},
            ],
            response_format=output_parser.pydantic_object,
        )
        result = completion.choices[0].message.parsed
        casted_result = cast(OutputType, result)
        return casted_result

    result = await chain.ainvoke(input_values, config={"callbacks": [logging_handler]})
    try:
        parsed_result = output_parser.invoke(result)
    except Exception as e:
        if isinstance(output_parser, ScriptOutputParser):
            raise e  # the problem has been handled in the parser
        log.debug(
            f"[red] Failed to parse result: {result}\nEncounter Exception {e}\nstart to reparse",
            extra={"markup": True},
        )
        reformat_parsed_result = format_bad_output(
            result,
            format_instructions=output_parser.get_format_instructions(),
            model_name=bad_output_process_model,
            use_fixed_model_version=use_fixed_model_version,
        )
        parsed_result = output_parser.invoke(reformat_parsed_result)
    log.info(f"Generated result: {parsed_result}")
    return parsed_result


@gin.configurable
@beartype
async def agenerate_env_profile(
    model_name: str,
    inspiration_prompt: str = "asking my boyfriend to stop being friends with his ex",
    examples: str = "",
    temperature: float = 0.7,
    bad_output_process_model: str = DEFAULT_BAD_OUTPUT_PROCESS_MODEL,
    use_fixed_model_version: bool = True,
) -> tuple[EnvironmentProfile, str]:
    """
    Using langchain to generate the background
    """
    return await agenerate(
        model_name=model_name,
        template="""Please generate scenarios and goals based on the examples below as well as the inspirational prompt, when creating the goals, try to find one point that both sides may not agree upon initially and need to collaboratively resolve it.
        Examples:
        {examples}
        Inspirational prompt: {inspiration_prompt}
        Please use the following format:
        {format_instructions}
        """,
        input_values=dict(
            inspiration_prompt=inspiration_prompt,
            examples=examples,
        ),
        output_parser=PydanticOutputParser(pydantic_object=EnvironmentProfile),
        temperature=temperature,
        bad_output_process_model=bad_output_process_model,
        use_fixed_model_version=use_fixed_model_version,
    )


@beartype
async def agenerate_relationship_profile(
    model_name: str,
    agents_profiles: list[str],
    bad_output_process_model: str = DEFAULT_BAD_OUTPUT_PROCESS_MODEL,
    use_fixed_model_version: bool = True,
) -> tuple[RelationshipProfile, str]:
    """
    Using langchain to generate the background
    """
    agent_profile = "\n".join(agents_profiles)
    return await agenerate(
        model_name=model_name,
        template="""Please generate relationship between two agents based on the agents' profiles below. Note that you generate
        {agent_profile}
        Please use the following format:
        {format_instructions}
        """,
        input_values=dict(
            agent_profile=agent_profile,
        ),
        output_parser=PydanticOutputParser(pydantic_object=RelationshipProfile),
        bad_output_process_model=bad_output_process_model,
        use_fixed_model_version=use_fixed_model_version,
    )


@gin.configurable
@beartype
async def agenerate_action(
    model_name: str,
    history: str,
    turn_number: int,
    action_types: list[ActionType],
    agent: str,
    goal: str,
    temperature: float = 0.7,
    script_like: bool = False,
    bad_output_process_model: str = DEFAULT_BAD_OUTPUT_PROCESS_MODEL,
    use_fixed_model_version: bool = True,
) -> AgentAction:
    """
    Using langchain to generate an example episode
    """
    try:
        if script_like:
            # model as playwright
            template = """
                Now you are a famous playwright, your task is to continue writing one turn for agent {agent} under a given background and history to help {agent} reach social goal. Please continue the script based on the previous turns. You can only generate one turn at a time.
                You can find {agent}'s background and goal in the 'Here is the context of the interaction' field.
                You should try your best to achieve {agent}'s goal in a way that align with their character traits.
                Additionally, maintaining the conversation's naturalness and realism is essential (e.g., do not repeat what other people has already said before).
                {history}.
                The script has proceeded to Turn #{turn_number}. Current available action types are
                {action_list}.
                Note: The script can be ended if 1. one agent have achieved social goals, 2. this conversation makes the agent uncomfortable, 3. the agent find it uninteresting/you lose your patience, 4. or for other reasons you think it should stop.

                Please only generate a JSON string including the action type and the argument.
                Your action should follow the given format:
                {format_instructions}
            """
        else:
            # Normal case, model as agent
            template = """
                Imagine you are {agent}, your task is to act/speak as {agent} would, keeping in mind {agent}'s social goal.
                You can find {agent}'s goal (or background) in the 'Here is the context of the interaction' field.
                Note that {agent}'s goal is only visible to you.
                You should try your best to achieve {agent}'s goal in a way that align with their character traits.
                Additionally, maintaining the conversation's naturalness and realism is essential (e.g., do not repeat what other people has already said before).
                {history}.
                You are at Turn #{turn_number}. Your available action types are
                {action_list}.
                Note: You can "leave" this conversation if 1. you have achieved your social goals, 2. this conversation makes you uncomfortable, 3. you find it uninteresting/you lose your patience, 4. or for other reasons you want to leave.

                Please only generate a JSON string including the action type and the argument.
                Your action should follow the given format:
                {format_instructions}
            """
        return await agenerate(
            model_name=model_name,
            template=template,
            input_values=dict(
                agent=agent,
                turn_number=str(turn_number),
                history=history,
                action_list=" ".join(action_types),
            ),
            output_parser=PydanticOutputParser(pydantic_object=AgentAction),
            temperature=temperature,
            bad_output_process_model=bad_output_process_model,
            use_fixed_model_version=use_fixed_model_version,
        )
    except Exception:
        return AgentAction(action_type="none", argument="")


@gin.configurable
@beartype
async def agenerate_script(
    model_name: str,
    background: ScriptBackground,
    temperature: float = 0.7,
    agent_names: list[str] = [],
    agent_name: str = "",
    history: str = "",
    single_step: bool = False,
    bad_output_process_model: str = DEFAULT_BAD_OUTPUT_PROCESS_MODEL,
    use_fixed_model_version: bool = True,
) -> tuple[ScriptInteractionReturnType, str]:
    """
    Using langchain to generate an the script interactions between two agent
    The script interaction is generated in a single generation process.
    Note that in this case we do not require a json format response,
    so the failure rate will be higher, and it is recommended to use at least llama-2-70b.
    """
    try:
        if single_step:
            return await agenerate(
                model_name=model_name,
                template="""Now you are a famous playwright, your task is to continue writing one turn for agent {agent} under a given background and history to help {agent} reach social goal. Please continue the script based on the previous turns. You can only generate one turn at a time.

                Here are the conversation background and history:
                {background}
                {history}

                Remember that you are an independent scriptwriter and should finish the script by yourself.
                The output should only contain the script following the format instructions, with no additional comments or text.

                Here are the format instructions:
                {format_instructions}""",
                input_values=dict(
                    background=background.to_natural_language(),
                    history=history,
                    agent=agent_name,
                ),
                output_parser=ScriptOutputParser(  # type: ignore[arg-type]
                    agent_names=agent_names,
                    background=background.to_natural_language(),
                    single_turn=True,
                ),
                temperature=temperature,
                bad_output_process_model=bad_output_process_model,
                use_fixed_model_version=use_fixed_model_version,
            )

        else:
            return await agenerate(
                model_name=model_name,
                template="""
                Please write the script between two characters based on their social goals with a maximum of 20 turns.

                {background}
                Your action should follow the given format:
                {format_instructions}
                Remember that you are an independent scriptwriter and should finish the script by yourself.
                The output should only contain the script following the format instructions, with no additional comments or text.""",
                input_values=dict(
                    background=background.to_natural_language(),
                ),
                output_parser=ScriptOutputParser(  # type: ignore[arg-type]
                    agent_names=agent_names,
                    background=background.to_natural_language(),
                    single_turn=False,
                ),
                temperature=temperature,
                bad_output_process_model=bad_output_process_model,
                use_fixed_model_version=use_fixed_model_version,
            )
    except Exception as e:
        # TODO raise(e) # Maybe we do not want to return anything?
        print(f"Exception in agenerate {e}")
        return_default_value: ScriptInteractionReturnType = (
            ScriptInteraction.default_value_for_return_type()
        )
        return (return_default_value, "")


@beartype
def process_history(
    script: ScriptBackground | EnvResponse | dict[str, AgentAction],
) -> str:
    """
    Format the script background
    """
    result = ""
    if isinstance(script, ScriptBackground | EnvResponse):
        script = script.dict()
        result = "The initial observation\n\n"
    for key, value in script.items():
        if value:
            result += f"{key}: {value} \n"
    return result


@beartype
async def agenerate_init_profile(
    model_name: str,
    basic_info: dict[str, str],
    bad_output_process_model: str = DEFAULT_BAD_OUTPUT_PROCESS_MODEL,
    use_fixed_model_version: bool = True,
) -> str:
    """
    Using langchain to generate the background
    """
    return await agenerate(
        model_name=model_name,
        template="""Please expand a fictional background for {name}. Here is the basic information:
            {name}'s age: {age}
            {name}'s gender identity: {gender_identity}
            {name}'s pronouns: {pronoun}
            {name}'s occupation: {occupation}
            {name}'s big 5 personality traits: {bigfive}
            {name}'s moral Foundation: think {mft} is more important than others
            {name}'s Schwartz portrait value: {schwartz}
            {name}'s decision-making style: {decision_style}
            {name}'s secret: {secret}
            Include the previous information in the background.
            Then expand the personal backgrounds with concrete details (e.g, look, family, hobbies, friends and etc.)
            For the personality and values (e.g., MBTI, moral foundation, and etc.),
            remember to use examples and behaviors in the person's life to demonstrate it.
            """,
        input_values=dict(
            name=basic_info["name"],
            age=basic_info["age"],
            gender_identity=basic_info["gender_identity"],
            pronoun=basic_info["pronoun"],
            occupation=basic_info["occupation"],
            bigfive=basic_info["Big_Five_Personality"],
            mft=basic_info["Moral_Foundation"],
            schwartz=basic_info["Schwartz_Portrait_Value"],
            decision_style=basic_info["Decision_making_Style"],
            secret=basic_info["secret"],
        ),
        output_parser=StrOutputParser(),
        bad_output_process_model=bad_output_process_model,
        use_fixed_model_version=use_fixed_model_version,
    )


@beartype
async def convert_narratives(
    model_name: str,
    narrative: str,
    text: str,
    bad_output_process_model: str = DEFAULT_BAD_OUTPUT_PROCESS_MODEL,
    use_fixed_model_version: bool = True,
) -> str:
    if narrative == "first":
        return await agenerate(
            model_name=model_name,
            template="""Please convert the following text into a first-person narrative.
            e.g, replace name, he, she, him, her, his, and hers with I, me, my, and mine.
            {text}""",
            input_values=dict(text=text),
            output_parser=StrOutputParser(),
            bad_output_process_model=bad_output_process_model,
            use_fixed_model_version=use_fixed_model_version,
        )
    elif narrative == "second":
        return await agenerate(
            model_name=model_name,
            template="""Please convert the following text into a second-person narrative.
            e.g, replace name, he, she, him, her, his, and hers with you, your, and yours.
            {text}""",
            input_values=dict(text=text),
            output_parser=StrOutputParser(),
            bad_output_process_model=bad_output_process_model,
            use_fixed_model_version=use_fixed_model_version,
        )
    else:
        raise ValueError(f"Narrative {narrative} is not supported.")


@beartype
async def agenerate_goal(
    model_name: str,
    background: str,
    bad_output_process_model: str = DEFAULT_BAD_OUTPUT_PROCESS_MODEL,
    use_fixed_model_version: bool = True,
) -> str:
    """
    Using langchain to generate the background
    """
    return await agenerate(
        model_name=model_name,
        template="""Please generate your goal based on the background:
            {background}
            """,
        input_values=dict(background=background),
        output_parser=StrOutputParser(),
        bad_output_process_model=bad_output_process_model,
        use_fixed_model_version=use_fixed_model_version,
    )



from pydantic import validator
from redis_om import EmbeddedJsonModel, JsonModel
from redis_om.model.model import Field

from .auto_expires_mixin import AutoExpireMixin


class MessageTransaction(EmbeddedJsonModel):
    timestamp_str: str = Field(index=True)
    sender: str = Field(index=True)
    message: str

    def to_tuple(self) -> tuple[float, str, str]:
        return (
            float(self.timestamp_str),
            self.sender,
            self.message,
        )


class SessionTransaction(AutoExpireMixin, JsonModel):
    session_id: str = Field(index=True)
    client_id: str = Field(index=True)
    server_id: str = Field(index=True)
    client_action_lock: str = Field(default="no action")
    message_list: list[MessageTransaction] = Field(
        description="""List of messages in this session.
    Each message is a tuple of (timestamp, sender_id, message)
    The message list should be sorted by timestamp.
    """
    )

    @validator("message_list")
    def validate_message_list(
        cls, v: list[MessageTransaction]
    ) -> list[MessageTransaction]:
        def _is_sorted(x: list[MessageTransaction]) -> bool:
            return all(
                float(x[i].timestamp_str) <= float(x[i + 1].timestamp_str)
                for i in range(len(x) - 1)
            )

        assert _is_sorted(v), "Message list should be sorted by timestamp"
        return v



from typing import Generic, TypeVar

from redis_om.model.model import NotFoundError

from sotopia.database import AgentProfile
from sotopia.messages import MessengerMixin

ObsType = TypeVar("ObsType")
ActType = TypeVar("ActType")


class BaseAgent(Generic[ObsType, ActType], MessengerMixin):
    def __init__(
        self,
        agent_name: str | None = None,
        uuid_str: str | None = None,
        agent_profile: AgentProfile | None = None,
    ) -> None:
        MessengerMixin.__init__(self)
        if agent_profile is not None:
            self.profile = agent_profile
            self.agent_name = self.profile.first_name + " " + self.profile.last_name
        elif uuid_str is not None:
            # try retrieving profile from database
            try:
                self.profile = AgentProfile.get(pk=uuid_str)
            except NotFoundError:
                raise ValueError(f"Agent with uuid {uuid_str} not found in database")
            self.agent_name = self.profile.first_name + " " + self.profile.last_name
        else:
            assert (
                agent_name is not None
            ), "Either agent_name or uuid_str must be provided"
            self.agent_name = agent_name

        self._goal: str | None = None
        self.model_name: str = ""

    @property
    def goal(self) -> str:
        assert self._goal is not None, "attribute goal has to be set before use"
        return self._goal

    @goal.setter
    def goal(self, goal: str) -> None:
        self._goal = goal

    def act(self, obs: ObsType) -> ActType:
        raise NotImplementedError

    async def aact(self, obs: ObsType) -> ActType:
        raise NotImplementedError

    def reset(self) -> None:
        self.reset_inbox()



from typing import Any, Generator, Generic, Sequence, Type, TypeVar

from sotopia.agents.base_agent import BaseAgent
from sotopia.database.persistent_profile import (
    AgentProfile,
    EnvironmentProfile,
)
from sotopia.envs.parallel import ParallelSotopiaEnv

ObsType = TypeVar("ObsType")
ActType = TypeVar("ActType")

EnvAgentCombo = tuple[ParallelSotopiaEnv, Sequence[BaseAgent[ObsType, ActType]]]


class BaseSampler(Generic[ObsType, ActType]):
    def __init__(
        self,
        env_candidates: Sequence[EnvironmentProfile | str] | None = None,
        agent_candidates: Sequence[AgentProfile | str] | None = None,
    ) -> None:
        self.env_candidates = env_candidates
        self.agent_candidates = agent_candidates

    def sample(
        self,
        agent_classes: Type[BaseAgent[ObsType, ActType]]
        | list[Type[BaseAgent[ObsType, ActType]]],
        n_agent: int = 2,
        replacement: bool = True,
        size: int = 1,
        env_params: dict[str, Any] = {},
        agents_params: list[dict[str, Any]] = [{}, {}],
    ) -> Generator[EnvAgentCombo[ObsType, ActType], None, None]:
        """Sample an environment and a list of agents.

        Args:
            agent_classes (Type[BaseAgent[ObsType, ActType]] | list[Type[BaseAgent[ObsType, ActType]]]): A single agent class for all sampled agents or a list of agent classes.
            n_agent (int, optional): Number of agents. Defaults to 2.
            replacement (bool, optional): Whether to sample with replacement. Defaults to True.
            size (int, optional): Number of samples. Defaults to 1.
            env_params (dict[str, Any], optional): Parameters for the environment. Defaults to {}.
            agents_params (list[dict[str, Any]], optional): Parameters for the agents. Defaults to [{}, {}].

        Returns:
            tuple[ParallelSotopiaEnv, list[BaseAgent[ObsType, ActType]]]: an environment and a list of agents.
        """
        raise NotImplementedError



from __future__ import annotations

import logging
from typing import Any, Dict, List, Mapping, Optional, Tuple, cast

import together
from langchain.callbacks.manager import (
    AsyncCallbackManagerForLLMRun,
    CallbackManagerForLLMRun,
)
from langchain.chat_models.base import BaseChatModel
from langchain.schema import (
    AIMessage,
    BaseMessage,
    ChatGeneration,
    ChatMessage,
    ChatResult,
    HumanMessage,
    SystemMessage,
)
from langchain.utils import get_from_dict_or_env
from pydantic import Extra, root_validator


def _convert_message_to_dict(message: BaseMessage) -> dict[str, Any]:
    if isinstance(message, ChatMessage):
        message_dict = {"role": message.role, "content": message.content}
    elif isinstance(message, HumanMessage):
        message_dict = {"role": "user", "content": message.content}
    elif isinstance(message, AIMessage):
        message_dict = {"role": "assistant", "content": message.content}
    elif isinstance(message, SystemMessage):
        message_dict = {"role": "system", "content": message.content}
    else:
        raise ValueError(f"Got unknown type {message}")
    if "name" in message.additional_kwargs:
        message_dict["name"] = message.additional_kwargs["name"]
    return message_dict


def _convert_dict_to_message(_dict: dict[str, str]) -> BaseMessage:
    text = _dict["text"]
    return AIMessage(content=text)


def _make_prompt_from_dict(dialog: List[dict[str, str]]) -> str:
    """
    Follow chat format in https://docs.together.ai/docs/python-chat
    example: together complete "<human>: List the best restaurants in SF\n<bot>: "
    python example: together.Complete.create(prompt=(above), model=model_name)
    The convertion is similar to llama2 official, EXCEPT not using special tags:
    https://github.com/facebookresearch/llama/blob/main/example_chat_completion.py
    """
    user_tag = "<human>"
    assistant_tag = "<bot>"
    tag_sep = ": "
    conv_sep = "\n"

    if dialog[0]["role"] == "system":
        dialog = [
            {
                "role": dialog[1]["role"],
                "content": dialog[0]["content"] + conv_sep + dialog[1]["content"],
            }
        ] + dialog[2:]
    assert all([msg["role"] == "user" for msg in dialog[::2]]) and all(
        [msg["role"] == "assistant" for msg in dialog[1::2]]
    ), (
        "model only supports 'system', 'user' and 'assistant' roles, "
        "starting with 'system', then 'user' and alternating (u/a/u/a/u...)"
    )
    ret = ""
    for d in dialog:
        tt = user_tag
        if d["role"] == "user":
            tt = user_tag
        elif d["role"] == "assistant":
            tt = assistant_tag
        ret += tt + tag_sep + d["content"].strip() + conv_sep
    ret += assistant_tag + tag_sep
    return ret


logger = logging.getLogger(__name__)


class Llama2(BaseChatModel):
    client: type[together.Complete] = together.Complete  #: :meta private:
    model_name: str = "togethercomputer/llama-2-7b-chat"
    """Model name to use."""
    # default Together params
    temperature: float = 0.7
    max_tokens: int = 1024
    top_p: float = 0.7
    top_k: int = 50
    repetition_penalty: float = 1.0
    start: bool = False
    _llm_type: str = "llama2"

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.ignore

    @root_validator(pre=True)
    def build_extra(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Build extra kwargs from additional params that were passed in."""
        all_required_field_names = {field.alias for field in cls.__fields__.values()}

        extra = values.get("model_kwargs", {})
        for field_name in list(values):
            if field_name not in all_required_field_names:
                if field_name in extra:
                    raise ValueError(f"Found {field_name} supplied twice.")
                extra[field_name] = values.pop(field_name)
        values["model_kwargs"] = extra
        return values

    @root_validator()
    def validate_environment(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        together_api_key = get_from_dict_or_env(
            values, "together_api_key", "TOGETHER_API_KEY"
        )
        together.api_key = together_api_key
        values["client"] = together.Complete
        return values

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        prompt, params = self._create_message_dicts(messages, stop)
        response = self.client.create(prompt=prompt, **params)
        chat_result = self._create_chat_result(response)
        return chat_result

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling Together API."""
        return {
            "model": self.model_name,
            "max_tokens": self.max_tokens,
            "temperature": self.temperature,
            "top_p": self.top_p,
            "top_k": self.top_k,
            "repetition_penalty": self.repetition_penalty,
        }

    def _create_message_dicts(
        self, messages: List[BaseMessage], stop: Optional[List[str]]
    ) -> Tuple[str, Dict[str, Any]]:
        params: Dict[str, Any] = {
            **{"model": self.model_name},
            **self._default_params,
        }
        if stop is not None:
            if "stop" in params:
                raise ValueError("`stop` found in both the input and default params.")
            params["stop"] = stop
        message_dicts = [_convert_message_to_dict(m) for m in messages]
        message_prompt = _make_prompt_from_dict(message_dicts)
        return message_prompt, params

    def _create_chat_result(self, response: Mapping[str, Any]) -> ChatResult:
        generations = []
        for res in response["output"]["choices"]:
            message = _convert_dict_to_message(res)
            gen = ChatGeneration(message=message)
            generations.append(gen)
        llm_output = {"model_name": self.model_name}
        return ChatResult(generations=generations, llm_output=llm_output)

    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: List[str] | None = None,
        run_manager: AsyncCallbackManagerForLLMRun | None = None,
        **kwargs: Any,
    ) -> ChatResult:
        sync_run_manager = cast(CallbackManagerForLLMRun, run_manager)
        return self._generate(messages, stop, sync_run_manager)



import asyncio
import logging
import os
from uuid import uuid4

import aiohttp
import pydantic
import requests

from sotopia.agents import BaseAgent
from sotopia.database import AgentProfile, MessageTransaction
from sotopia.generation_utils.generate import LLM_Name
from sotopia.messages import AgentAction, Observation


class RedisAgent(BaseAgent[Observation, AgentAction]):
    """An agent use redis as a message broker."""

    def __init__(
        self,
        agent_name: str | None = None,
        uuid_str: str | None = None,
        session_id: str | None = None,
        agent_profile: AgentProfile | None = None,
    ) -> None:
        super().__init__(
            agent_name=agent_name,
            uuid_str=uuid_str,
            agent_profile=agent_profile,
        )
        # super().__init__(agent_name=agent_name, uuid_str=uuid_str)
        self.session_id = session_id or str(uuid4())
        self.sender_id = str(uuid4())
        self.model_name: LLM_Name = "redis"
        print(f"session id: {self.session_id}")
        print("step 1: connect to the server")
        assert (
            "FASTAPI_URL" in os.environ
        ), "To use redis agent, you have to launch a FastAPI server and set FASTAPI_URL"
        self._URL = os.environ["FASTAPI_URL"]
        response = requests.request(
            "POST",
            f"{self._URL}/connect/{self.session_id}/server/{self.sender_id}",
        )
        assert (
            response.status_code == 200 and response.text == "[]"
        ), "Failed to connect to the server"
        logging.info(f"Session ID: {self.session_id}")
        # logging.info(f"Sender ID: {self.sender_id}")

    def act(
        self,
        obs: Observation,
    ) -> AgentAction:
        raise NotImplementedError

    async def aact(
        self,
        obs: Observation,
    ) -> AgentAction:
        self.recv_message("Environment", obs)

        if len(obs.available_actions) == 1 and "none" in obs.available_actions:
            if obs.turn_number == 0:
                async with aiohttp.ClientSession() as session:
                    print("step 2: post observation to the message list")
                    response = await session.request(
                        "POST",
                        f"{self._URL}/send/{self.session_id}/{self.sender_id}",
                        data=obs.to_natural_language(),
                    )
                    assert response.status == 200, response
                    sorted_message_list: list[tuple[float, str, str]] = list(
                        map(
                            lambda x: MessageTransaction.parse_obj(x).to_tuple(),
                            await response.json(),
                        )
                    )
                    last_timestamp = sorted_message_list[-1][0]
            return AgentAction(action_type="none", argument="")
        else:
            async with aiohttp.ClientSession() as session:
                # 1. post observation to the message list
                response = await session.request(
                    "POST",
                    f"{self._URL}/send/{self.session_id}/{self.sender_id}",
                    data=obs.to_natural_language(),
                )
                assert response.status == 200, response
                sorted_message_list = list(
                    map(
                        lambda x: MessageTransaction.parse_obj(x).to_tuple(),
                        await response.json(),
                    )
                )
                last_timestamp = sorted_message_list[-1][0]

                print("step 2: unlock the server for the client")
                # 2. unlock the server for the client
                response = await session.request(
                    "PUT",
                    f"{self._URL}/lock/{self.session_id}/{self.sender_id}/action",
                )
                assert response.status == 200, response

                print("step 3: wait for the client to post their message")
                # 3. wait for the client to post their message
                for _ in range(300):
                    response = await session.request(
                        "GET",
                        f"{self._URL}/get/{self.session_id}",
                    )
                    # print(f"get response: {response}")
                    assert response.status == 200, response
                    sorted_message_list = list(
                        map(
                            lambda x: MessageTransaction.parse_obj(x).to_tuple(),
                            await response.json(),
                        )
                    )
                    if (
                        sorted_message_list[-1][0] > last_timestamp
                        and sorted_message_list[-1][1] == "client"
                    ):
                        # 3.a if the client has posted their message, lock the server for the client
                        response = await session.request(
                            "PUT",
                            f"{self._URL}/lock/{self.session_id}/{self.sender_id}/no%20action",
                        )
                        assert response.status == 200, response
                        break
                    else:
                        # 3.b if the client has not posted their message, wait for 0.1 second and retry
                        await asyncio.sleep(1)
                else:
                    response = await session.request(
                        "PUT",
                        f"{self._URL}/lock/{self.session_id}/{self.sender_id}/no%20action",
                    )
                    self.reset("Someone has left or the conversation is too long.")
                    return AgentAction(action_type="leave", argument="")
            action_string = sorted_message_list[-1][2]
            try:
                action = AgentAction.parse_raw(action_string)
                return action
            except pydantic.error_wrappers.ValidationError:
                logging.warn(
                    "Failed to parse action string {}. Fall back to speak".format(
                        action_string
                    )
                )
                return AgentAction(
                    action_type="speak", argument=sorted_message_list[-1][2]
                )

    def reset(
        self,
        reset_reason: str = "",
    ) -> None:
        super().reset()
        try:
            if reset_reason != "":
                response = requests.request(
                    "POST",
                    f"{self._URL}/send/{self.session_id}/{self.sender_id}",
                    json=reset_reason,
                )
                assert response.status_code == 200

        except Exception as e:
            logging.error(f"Failed to reset RedisAgent {self.sender_id}: {e}")



import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import cast

from sotopia.agents import BaseAgent
from sotopia.database import AgentProfile
from sotopia.generation_utils.generate import (
    LLM_Name,
    agenerate_action,
    agenerate_goal,
    agenerate_script,
)
from sotopia.messages import AgentAction, Observation
from sotopia.messages.message_classes import ScriptBackground


async def ainput(prompt: str = "") -> str:
    with ThreadPoolExecutor(1, "ainput") as executor:
        return (
            await asyncio.get_event_loop().run_in_executor(executor, input, prompt)
        ).rstrip()


class LLMAgent(BaseAgent[Observation, AgentAction]):
    def __init__(
        self,
        agent_name: str | None = None,
        uuid_str: str | None = None,
        agent_profile: AgentProfile | None = None,
        model_name: str = "gpt-4o-mini",
        script_like: bool = False,
    ) -> None:
        super().__init__(
            agent_name=agent_name,
            uuid_str=uuid_str,
            agent_profile=agent_profile,
        )
        self.model_name = model_name
        self.script_like = script_like

    @property
    def goal(self) -> str:
        if self._goal is not None:
            return self._goal
        else:
            raise Exception("Goal is not set.")

    @goal.setter
    def goal(self, goal: str) -> None:
        self._goal = goal

    def act(
        self,
        _obs: Observation,
    ) -> AgentAction:
        raise Exception("Sync act method is deprecated. Use aact instead.")

    async def aact(self, obs: Observation) -> AgentAction:
        self.recv_message("Environment", obs)

        if self._goal is None:
            self._goal = await agenerate_goal(
                self.model_name,
                background=self.inbox[0][
                    1
                ].to_natural_language(),  # Only consider the first message for now
            )

        if len(obs.available_actions) == 1 and "none" in obs.available_actions:
            return AgentAction(action_type="none", argument="")
        else:
            action = await agenerate_action(
                self.model_name,
                history="\n".join(f"{y.to_natural_language()}" for x, y in self.inbox),
                turn_number=obs.turn_number,
                action_types=obs.available_actions,
                agent=self.agent_name,
                goal=self.goal,
                script_like=self.script_like,
            )
            # Temporary fix for mixtral-moe model for incorrect generation format
            if "Mixtral-8x7B-Instruct-v0.1" in self.model_name:
                current_agent = self.agent_name
                if f"{current_agent}:" in action.argument:
                    print("Fixing Mixtral's generation format")
                    action.argument = action.argument.replace(f"{current_agent}: ", "")
                elif f"{current_agent} said:" in action.argument:
                    print("Fixing Mixtral's generation format")
                    action.argument = action.argument.replace(
                        f"{current_agent} said: ", ""
                    )

            return action


class ScriptWritingAgent(LLMAgent):
    def __init__(
        self,
        agent_name: str | None = None,
        uuid_str: str | None = None,
        agent_profile: AgentProfile | None = None,
        model_name: str = "gpt-4o-mini",
        agent_names: list[str] = [],
        background: ScriptBackground | None = None,
    ) -> None:
        super().__init__(
            agent_name=agent_name,
            uuid_str=uuid_str,
            agent_profile=agent_profile,
        )
        self.model_name = model_name
        self.agent_names = agent_names
        assert background is not None, "background cannot be None"
        self.background = background

    async def aact(self, obs: Observation) -> AgentAction:
        self.recv_message("Environment", obs)
        message_to_compose = [y for idx, (x, y) in enumerate(self.inbox) if idx != 0]

        history = "\n".join(f"{y.to_natural_language()}" for y in message_to_compose)

        action, prompt = await agenerate_script(
            model_name=self.model_name,
            background=self.background,
            agent_names=self.agent_names,
            history=history,
            agent_name=self.agent_name,
            single_step=True,
        )
        returned_action = cast(AgentAction, action[1][0][1])
        return returned_action


class HumanAgent(BaseAgent[Observation, AgentAction]):
    """
    A human agent that takes input from the command line.
    """

    def __init__(
        self,
        agent_name: str | None = None,
        uuid_str: str | None = None,
        agent_profile: AgentProfile | None = None,
    ) -> None:
        super().__init__(
            agent_name=agent_name,
            uuid_str=uuid_str,
            agent_profile=agent_profile,
        )
        self.model_name: LLM_Name = "human"

    @property
    def goal(self) -> str:
        if self._goal is not None:
            return self._goal
        goal = input("Goal: ")
        return goal

    @goal.setter
    def goal(self, goal: str) -> None:
        self._goal = goal

    def act(self, obs: Observation) -> AgentAction:
        self.recv_message("Environment", obs)

        print("Available actions:")
        for i, action in enumerate(obs.available_actions):
            print(f"{i}: {action}")

        action_type = obs.available_actions[int(input("Action type: "))]
        argument = input("Argument: ")

        return AgentAction(action_type=action_type, argument=argument)

    async def aact(self, obs: Observation) -> AgentAction:
        self.recv_message("Environment", obs)

        print("Available actions:")
        for i, action in enumerate(obs.available_actions):
            print(f"{i}: {action}")

        if obs.available_actions != ["none"]:
            action_type_number = await ainput(
                "Action type (Please only input the number): "
            )
            try:
                action_type_number = int(action_type_number)  # type: ignore
            except TypeError:
                print("Please input a number.")
                action_type_number = await ainput(
                    "Action type (Please only input the number): "
                )
                action_type_number = int(action_type_number)  # type: ignore
            assert isinstance(action_type_number, int), "Please input a number."
            action_type = obs.available_actions[action_type_number]
        else:
            action_type = "none"
        if action_type in ["speak", "non-verbal communication"]:
            argument = await ainput("Argument: ")
        else:
            argument = ""

        return AgentAction(action_type=action_type, argument=argument)


class Agents(dict[str, BaseAgent[Observation, AgentAction]]):
    def reset(self) -> None:
        for agent in self.values():
            agent.reset()

    def act(self, obs: dict[str, Observation]) -> dict[str, AgentAction]:
        return {
            agent_name: agent.act(obs[agent_name]) for agent_name, agent in self.items()
        }
